import pprint

import simgen.parse_makefile as SPM
import simgen.basic_simulator as SBS
import simgen.ibm1130_simulator as IBM1130
import simgen.pdp10_simulator as PDP10
import simgen.vax_simulators as VAXen
import simgen.utils as SU

## Special variables that should __not__ expand into their definitions:
_special_vars = frozenset(['DISPLAYL',
    'DISPLAYVT',
    'DISPLAY340',
    'DISPLAYNG',
    'DISPLAYIII'])

## Map simulator name to its class, for special cases
_special_simulators = {
    "besm6": SBS.BESM6Simulator,
    "i650": SBS.IBM650Simulator,
    "ibm1130": IBM1130.IBM1130Simulator,
    "pdp10-ka": PDP10.KA10Simulator,
    "vax": VAXen.VAXSimulator,
    "vax730": VAXen.BasicVAXSimulator
}

ignored_display_macros = {
    'DISPLAYVT': ['${DISPLAYD}/vt11.c'],
    'DISPLAY340': ['${DISPLAYD}/type340.c'],
    'DISPLAYNG': ['${DISPLAYD}/ng.c'],
    'DISPLAYIII': ['${DISPLAYD}/iii.c']
}

def get_simulator_ctor(name):
    """Return the class object for special case simulators, otherwise
    return the base 'SIMHBasicSimulator'
    """
    return _special_simulators.get(name) or SBS.SIMHBasicSimulator


class SimCollection:
    """A collection of simulators.
    """
    def __init__(self, dir_macro):
        self.source_macros = {}
        self.macro_uses = {}
        self.simulators = {}

    def get_simulator(self, name, dir_macro, _dir_path, test_name, buildrom):
        sim = self.simulators.get(name)
        if sim is None:
            sim = (get_simulator_ctor(name))(name, dir_macro, test_name, buildrom)
            self.simulators[name] = sim
        return sim

    def add_source_macro(self, macro, macro_def, sim):
        if macro not in self.source_macros:
            self.source_macros[macro] = macro_def

        used = self.macro_uses.get(macro)
        if used is None:
            self.macro_uses[macro] = []
            used = self.macro_uses[macro]
        used.append(sim)

    def get_simulator_vars(self, debug=0):
        simvars = set()
        ignored = set(self.source_macros.keys())
        for macval in self.source_macros.values():
            ## This could be replaced by a functools.reduce()
            for val in macval:
                simvars = simvars.union(set(SPM.extract_variables(val)))

        for sim in self.simulators.values():
            simvars = simvars.union(sim.get_source_vars().union(sim.get_include_vars()))

        simvars = simvars.difference(ignored).difference(_special_vars)
        SU.emit_debug(debug, 2, 'simvars {0}'.format(simvars))
        return simvars

    def write_simulators(self, stream, debug=0, test_label='default'):
        ## Emit source macros
        dontexpand = set([smac for smac, uses in self.macro_uses.items() if smac not in ignored_display_macros and len(uses) > 1])
        SU.emit_debug(debug, 2, "{0}: dontexpand {1}".format(self.__class__.__name__, dontexpand))

        if len(dontexpand) > 0:
            smac_sorted = list(dontexpand)
            smac_sorted.sort()
            for smac in smac_sorted:
                stream.write('\n\n')
                stream.write('set({0}\n'.format(smac))
                stream.write('\n'.join([' ' * 4 + f for f in self.source_macros[smac]]))
                stream.write(')')
            stream.write('\n\n')

        ## Emit the simulators
        simnames = list(self.simulators.keys())
        simnames.sort()
        SU.emit_debug(debug, 2, "{0}: Writing {1}".format(self.__class__.__name__, simnames))
        for simname in simnames:
            sim = self.simulators[simname]

            ## Patch up the simulator source lists, expanding macros that aren't
            ## in the macro sources:
            sim.sources = self.expand_sources(sim.sources, dontexpand, debug)

            stream.write('\n')
            sim.write_simulator(stream, 0, test_label)

    def write_unit_tests(self, stream, debug=0, test_label='default'):
        dontexpand = set([smac for smac, uses in self.macro_uses.items() if len(uses) > 1])

        simnames = list(self.simulators.keys())
        simnames.sort()
        SU.emit_debug(debug, 2, "{0}: Writing {1}".format(self.__class__.__name__, simnames))
        for simname in simnames:
            sim = self.simulators[simname]

            ## Patch up the simulator source lists, expanding macros that aren't
            ## in the macro sources:
            sim.sources = self.expand_sources(sim.sources, dontexpand, debug)
            sim.write_unit_test(stream, 0, test_label)

    def expand_sources(self, srcs, dontexpand, debug=0):
        updated_srcs = []
        for src in srcs:
            SU.emit_debug(debug, 2, "{0}: Source {1}".format(self.__class__.__name__, src))
            m = SPM._var_rx.match(src)
            if m and m[1] not in dontexpand.union(_special_vars):
                SU.emit_debug(debug, 2, "{0}: Expanding {1}".format(self.__class__.__name__, m[1]))
                varexp = self.source_macros.get(m[1])
                if varexp is not None:
                    updated_srcs.extend(self.source_macros[m[1]])
                else:
                    print('!! Could not expand {0}'.format(m[1]))
            else:
                updated_srcs.append(src)

        if updated_srcs == srcs:
            return srcs
        else:
            return self.expand_sources(updated_srcs, dontexpand, debug)

    def __len__(self):
        return len(self.simulators)

if '_dispatch' in pprint.PrettyPrinter.__dict__:
    def simcoll_pprinter(pprinter, simcoll, stream, indent, allowance, context, level):
        cls = simcoll.__class__
        stream.write(cls.__name__ + '(')
        indent += len(cls.__name__) + 1
        pprinter._format(simcoll.source_macros, stream, indent, allowance + 2, context, level)
        stream.write(',\n' + ' ' * indent)
        uses_dict = dict([(sim, len(uses)) for (sim, uses) in simcoll.macro_uses.items()])
        pprinter._format(uses_dict, stream, indent, allowance + 2, context, level)
        stream.write(',\n' + ' ' * indent)
        pprinter._format(simcoll.simulators, stream, indent, allowance + 2, context, level)
        stream.write(')')

    pprint.PrettyPrinter._dispatch[SimCollection.__repr__] = simcoll_pprinter



import os
import functools

## Initialize package_info to an empty dictionary here so
## that it's visible to write_packaging().
package_info = {}


class SIMHPackaging:
    def __init__(self, family, install_flag = True) -> None:
        self.family = family
        self.processed = False
        self.install_flag = install_flag

    def was_processed(self) -> bool:
        return self.processed == True
    
    def encountered(self) -> None:
        self.processed = True

class PkgFamily:
    def __init__(self, component_name, display_name, description) -> None:
        self.component_name = component_name
        self.display_name   = display_name
        self.description    = description

    def write_component_info(self, stream, indent) -> None:
        pkg_description = self.description
        if pkg_description[-1] != '.':
            pkg_description += '.'
        sims = []
        for sim, pkg in package_info.items():
            if pkg.family is self and pkg.was_processed():
                sims.append(sim)

        if len(sims) > 0:
            sims.sort()
            pkg_description += " Simulators: " + ', '.join(sims)
            indent0 = ' ' * indent
            indent4 = ' ' * (indent + 4)
            stream.write(indent0 + "cpack_add_component(" + self.component_name + "\n")
            stream.write(indent4 + "DISPLAY_NAME \"" + self.display_name + "\"\n")
            stream.write(indent4 + "DESCRIPTION \"" + pkg_description + "\"\n")
            stream.write(indent0 + ")\n")

    def __lt__(self, obj):
        return self.component_name < obj.component_name
    def __eq__(self, obj):
        return self.component_name == obj.component_name
    def __gt__(self, obj):
        return self.component_name > obj.component_name
    def __hash__(self):
        return hash(self.component_name)

def write_packaging(toplevel_dir) -> None:
    families = set([sim.family for sim in package_info.values()])
    pkging_file = os.path.join(toplevel_dir, 'cmake', 'simh-packaging.cmake')
    print("==== writing {0}".format(pkging_file))
    with open(pkging_file, "w") as stream:
        ## Runtime support family:
        stream.write("""## The default runtime support component/family:
cpack_add_component(runtime_support
    DISPLAY_NAME "Runtime support"
    DESCRIPTION "Required SIMH runtime support (documentation, shared libraries)"
    REQUIRED
)

## Basic documentation for SIMH
install(FILES doc/simh.doc TYPE DOC COMPONENT runtime_support)

""")

        ## Simulators:
        for family in sorted(families):
            family.write_component_info(stream, 0)


default_family = PkgFamily("default_family", "Default SIMH simulator family.",
    """The SIMH simulator collection of historical processors and computing systems that do not belong to
any other simulated system family"""
)
    
att3b2_family = PkgFamily("att3b2_family", "ATT&T 3b2 collection",
    """The AT&T 3b2 simulator family"""
)

vax_family = PkgFamily("vax_family", "DEC VAX simulator collection",
    """The Digital Equipment Corporation VAX (plural: VAXen) simulator family."""
)

pdp10_family = PkgFamily("pdp10_family", "DEC PDP-10 collection",
    """DEC PDP-10 architecture simulators and variants."""
)

pdp11_family = PkgFamily("pdp11_family", "DEC PDP-11 collection.",
   """DEC PDP-11 and PDP-11-derived architecture simulators"""
)

experimental_family = PkgFamily("experimental", "Experimental (work-in-progress) simulators",
    """Experimental or work-in-progress simulators not in the SIMH mainline simulator suite."""
)

altairz80_family = PkgFamily("altairz80_family", "Altair Z80 simulator.",
    """The Altair Z80 simulator with M68000 support."""
)

b5500_family = PkgFamily("b5500_family", "Burroughs 5500",
    """The Burroughs 5500 system simulator""")

cdc1700_family = PkgFamily("cdc1700_family", "CDC 1700",
    """The Control Data Corporation's systems"""
)

dgnova_family = PkgFamily("dgnova_family", "DG Nova and Eclipse",
    """Data General NOVA and Eclipse system simulators"""
)

grisys_family = PkgFamily("grisys_family", "GRI Systems GRI-909",
    """GRI Systems GRI-909 system simulator"""
)

honeywell_family = PkgFamily("honeywell_family", "Honeywell H316",
    """Honeywell H-316 system simulator"""
)

hp_family = PkgFamily("hp_family", "HP 2100, 3000",
    """Hewlett-Packard H2100 and H3000 simulators""")

ibm_family = PkgFamily("ibm_family", "IBM",
    """IBM system simulators: i650"""
)

imlac_family = PkgFamily("imlac_family", "IMLAC",
    """IMLAC system simulators"""
)

intel_family = PkgFamily("intel_family", "Intel",
    """Intel system simulators"""
)

interdata_family = PkgFamily("interdata_family", "Interdata",
    """Interdata systems simulators"""
)

lgp_family = PkgFamily("lgp_family", "LGP",
    """Librascope systems"""
)

decpdp_family = PkgFamily("decpdp_family", "DEC PDP family",
    """Digital Equipment Corporation PDP systems"""
)

sds_family = PkgFamily("sds_family", "SDS simulators",
    """Scientific Data Systems (SDS) systems"""
)

gould_family = PkgFamily("gould_family", "Gould simulators",
    """Gould Systems simulators"""
)

swtp_family = PkgFamily("swtp_family", "SWTP simulators",
    """Southwest Technical Products (SWTP) system simulators"""
)

norsk_family = PkgFamily("norsk_family", "ND simulators",
    """Norsk Data systems simulator family""")


package_info["3b2"] = SIMHPackaging(att3b2_family)
package_info["3b2-700"] = SIMHPackaging(att3b2_family)
package_info["altair"] = SIMHPackaging(default_family)
package_info["altairz80"] = SIMHPackaging(altairz80_family)
package_info["b5500"] = SIMHPackaging(b5500_family)
package_info["besm6"] = SIMHPackaging(default_family)
package_info["cdc1700"] = SIMHPackaging(cdc1700_family)
package_info["eclipse"] = SIMHPackaging(dgnova_family)
package_info["gri"] = SIMHPackaging(grisys_family)
package_info["h316"] = SIMHPackaging(honeywell_family)
package_info["hp2100"] = SIMHPackaging(hp_family)
package_info["hp3000"] = SIMHPackaging(hp_family)
package_info["i1401"] = SIMHPackaging(ibm_family)
package_info["i1620"] = SIMHPackaging(ibm_family)
package_info["i650"] = SIMHPackaging(ibm_family)
package_info["i701"] = SIMHPackaging(ibm_family)
package_info["i7010"] = SIMHPackaging(ibm_family)
package_info["i704"] = SIMHPackaging(ibm_family)
package_info["i7070"] = SIMHPackaging(ibm_family)
package_info["i7080"] = SIMHPackaging(ibm_family)
package_info["i7090"] = SIMHPackaging(ibm_family)
package_info["i7094"] = SIMHPackaging(ibm_family)
package_info["ibm1130"] = SIMHPackaging(ibm_family)
package_info["id16"] = SIMHPackaging(interdata_family)
package_info["id32"] = SIMHPackaging(interdata_family)
package_info["imlac"] = SIMHPackaging(imlac_family)
package_info["infoserver100"] = SIMHPackaging(vax_family)
package_info["infoserver1000"] = SIMHPackaging(vax_family)
package_info["infoserver150vxt"] = SIMHPackaging(vax_family)
package_info["intel-mds"] = SIMHPackaging(intel_family)
package_info["lgp"] = SIMHPackaging(lgp_family)
package_info["microvax1"] = SIMHPackaging(vax_family)
package_info["microvax2"] = SIMHPackaging(vax_family)
package_info["microvax2000"] = SIMHPackaging(vax_family)
package_info["microvax3100"] = SIMHPackaging(vax_family)
package_info["microvax3100e"] = SIMHPackaging(vax_family)
package_info["microvax3100m80"] = SIMHPackaging(vax_family)
package_info["nd100"] = SIMHPackaging(norsk_family)
package_info["nova"] = SIMHPackaging(dgnova_family)
package_info["pdp1"] = SIMHPackaging(decpdp_family)
## Don't install pdp10 per Rob Cromwell
package_info["pdp10"] = SIMHPackaging(pdp10_family, install_flag=False)
package_info["pdp10-ka"] = SIMHPackaging(pdp10_family)
package_info["pdp10-ki"] = SIMHPackaging(pdp10_family)
package_info["pdp10-kl"] = SIMHPackaging(pdp10_family)
package_info["pdp10-ks"] = SIMHPackaging(pdp10_family)
package_info["pdp11"] = SIMHPackaging(pdp11_family)
package_info["pdp15"] = SIMHPackaging(decpdp_family)
package_info["pdp4"] = SIMHPackaging(decpdp_family)
package_info["pdp6"] = SIMHPackaging(decpdp_family)
package_info["pdp7"] = SIMHPackaging(decpdp_family)
package_info["pdp8"] = SIMHPackaging(decpdp_family)
package_info["pdp9"] = SIMHPackaging(decpdp_family)
package_info["rtvax1000"] = SIMHPackaging(vax_family)
package_info["s3"] = SIMHPackaging(ibm_family)
package_info["scelbi"] = SIMHPackaging(intel_family)
package_info["sds"] = SIMHPackaging(sds_family)
package_info["sel32"] = SIMHPackaging(gould_family)
package_info["sigma"] = SIMHPackaging(sds_family)
package_info["ssem"] = SIMHPackaging(default_family)
package_info["swtp6800mp-a"] = SIMHPackaging(swtp_family)
package_info["swtp6800mp-a2"] = SIMHPackaging(swtp_family)
package_info["tt2500"] = SIMHPackaging(default_family)
package_info["tx-0"] = SIMHPackaging(default_family)
package_info["uc15"] = SIMHPackaging(pdp11_family)
package_info["vax"] = SIMHPackaging(vax_family)
package_info["vax730"] = SIMHPackaging(vax_family)
package_info["vax750"] = SIMHPackaging(vax_family)
package_info["vax780"] = SIMHPackaging(vax_family)
package_info["vax8200"] = SIMHPackaging(vax_family)
package_info["vax8600"] = SIMHPackaging(vax_family)
package_info["vaxstation3100m30"] = SIMHPackaging(vax_family)
package_info["vaxstation3100m38"] = SIMHPackaging(vax_family)
package_info["vaxstation3100m76"] = SIMHPackaging(vax_family)
package_info["vaxstation4000m60"] = SIMHPackaging(vax_family)
package_info["vaxstation4000vlc"] = SIMHPackaging(vax_family)

## Experimental simulators:
package_info["alpha"] = SIMHPackaging(experimental_family)
package_info["pdq3"] = SIMHPackaging(experimental_family)
package_info["sage"] = SIMHPackaging(experimental_family)



## cmake_container.py
##
## A container for a collection of SIMH simulators
##
##

import sys
import os
import re
import pprint
import functools

import simgen.parse_makefile as SPM
import simgen.sim_collection as SC
import simgen.utils as SU

## Corresponding special variable uses in the makefile:
_special_sources = frozenset(['${DISPLAYL}', '$(DISPLAYL)'])

## Banner header for individual CMakeLists.txt files:
_individual_header = [
    '##',
    '## This is an automagically generated file. Do NOT EDIT.',
    '## Any changes you make will be overwritten!!',
    '##',
    '## Make changes to the SIMH top-level makefile and then run the',
    '## "cmake/generate.py" script to regenerate these files.',
    '##',
    '##     cd cmake; python -m generate --help',
    '##',
    '## ' + '-' * 60 + '\n'
    ]

## Banner header for individual unit test CMakeLists.txt files:
_unit_test_header = [
    '##',
    '## This is an automagically generated file. Do NOT EDIT.',
    '## Any changes you make will be overwritten!!',
    '##',
    '## If you need to make changes, modify write_unit_test()',
    '## method in the the underlying Python class in ',
    '## cmake/simgen/basic_simulator.py, then execute the',
    '## "cmake/generate.py" script to regenerate these files.',
    '##',
    '##     cd cmake; python -m generate --help',
    '##',
    '## ' + '-' * 60 + '\n'
    ]
## Unset the display variables when not building with video. A +10
## kludge.
_unset_display_vars = '''
if (NOT WITH_VIDEO)
    ### Hack: Unset these variables so that they don't expand if
    ### not building with video:
    set(DISPLAY340 "")
    set(DISPLAYIII "")
    set(DISPLAYNG  "")
    set(DISPLAYVT  "")
endif ()
'''

class CMakeBuildSystem:
    """A container for collections of SIMH simulators and automagic
    CMakeLists.txt driver.

    This is the top-level container that stores collections of SIMH simulators
    in the 'SIM'
    """

    ## Compile command line elements that are ignored.
    _ignore_compile_elems = frozenset(['${LDFLAGS}', '$(LDFLAGS)',
        '${CC_OUTSPEC}', '$(CC_OUTSPEC)',
        '${SIM}', '$(SIM)',
        '-o', '$@',
        '${NETWORK_OPT}', '$(NETWORK_OPT)',
        '${SCSI}', '$(SCSI)',
        '${DISPLAY_OPT}', '$(DISPLAY_OPT)',
        '${VIDEO_CCDEFS}', '$(VIDEO_CCDEFS)',
        '${VIDEO_LDFLAGS}', '$(VIDEO_LDFLAGS)',
        '${BESM6_PANEL_OPT}', '$(BESM6_PANEL_OPT)'])

    def __init__(self):
        # "Special" variables that we look for in source code lists and which we'll
        # emit into the CMakeLists.txt files.
        self.vars = SC.ignored_display_macros.copy()
        # Subdirectory -> SimCollection mapping
        self.dirs = {}


    def extract(self, compile_action, test_name, sim_dir, sim_name, defs, buildrom, debug=0, depth=''):
        """Extract sources, defines, includes and flags from the simulator's
        compile action in the makefile.
        """
        sim_dir_path = SPM.expand_vars(sim_dir, defs).replace('./', '')
        simcoll = self.dirs.get(sim_dir_path)
        if simcoll is None:
            simcoll = SC.SimCollection(sim_dir)
            self.dirs[sim_dir_path] = simcoll

        sim = simcoll.get_simulator(sim_name, sim_dir, sim_dir_path, test_name, buildrom)

        # Remove compile command line elements and do one level of variable expansion, split the resulting
        # string into a list.
        all_comps = []
        for comp in [ SPM.normalize_variables(act) for act in compile_action.split() if not self.compile_elems_to_ignore(act) ]:
            all_comps.extend(comp.split())

        if debug >= 3:
            print('{0}all_comps after filtering:'.format(depth))
            pprint.pp(all_comps)

        # Iterate through the final compile component list and extract source files, includes
        # and defines.
        #
        # Deferred: Looking for options that set the i64, z64, video library options.
        while all_comps:
            comp = all_comps[0]
            # print(':: comp {0}'.format(comp))
            if self.is_source(comp):
                # Source file...
                ## sim.add_source(comp.replace(sim_dir + '/', ''))
                sim.add_source(comp)
            elif comp.startswith('-I'):
                all_comps = self.process_flag(all_comps, defs, sim.add_include, depth)
            elif comp.startswith('-D'):
                all_comps = self.process_flag(all_comps, defs, sim.add_define, depth)
            elif comp.startswith('-L') or comp.startswith('-l'):
                ## It's a library path or library. Skip.
                pass
            else:
                # Solitary variable expansion?
                m = SPM._var_rx.match(comp)
                if m:
                    varname = m.group(1)
                    if varname not in SC._special_vars:
                        if not self.is_source_macro(comp, defs):
                            expand = [ SPM.normalize_variables(elem) for elem in SPM.shallow_expand_vars(comp, defs).split()
                                       if not self.source_elems_to_ignore(elem) ]
                            SU.emit_debug(debug, 3, '{0}var expanded {1} -> {2}'.format(depth, comp, expand))
                            all_comps[1:] = expand + all_comps[1:]
                        else:
                            ## Source macro
                            self.collect_source_macros(comp, defs, varname, simcoll, sim, debug, depth)
                    else:
                        sim.add_source(comp)
                else:
                    # Nope.
                    print('{0}unknown component: {1}'.format(depth, comp))
            all_comps = all_comps[1:]

        sim.scan_for_flags(defs)
        sim.cleanup_defines()

        if debug >= 2:
            pprint.pprint(sim)


    def compile_elems_to_ignore(self, elem):
        return (elem in self._ignore_compile_elems or elem.endswith('_LDFLAGS'))

    def source_elems_to_ignore(self, elem):
        return self.compile_elems_to_ignore(elem) or elem in _special_sources


    def is_source_macro(self, var, defs):
        """Is the macro/variable a list of sources?
        """
        expanded = SPM.expand_vars(var, defs).split()
        # print('is_source_macro {}'.format(expanded))
        return all(map(lambda src: self.is_source(src), expanded))


    def is_source(self, thing):
        return thing.endswith('.c')


    def process_flag(self, comps, defs, process_func, depth):
        if len(comps[0]) > 2:
            # "-Ddef"
            val = comps[0][2:]
        else:
            # "-D def"
            val = comps[1]
            comps = comps[1:]
        m = SPM._var_rx.match(val)
        if m:
            var = m.group(1)
            ## Gracefully deal with undefined variables (ATT3B2M400B2D is a good example)
            if var in defs:
                if var not in self.vars:
                    self.vars[var] = defs[var]
            else:
                print('{0}undefined make macro: {1}'.format(depth, var))

        process_func(val)
        return comps


    def collect_vars(self, defs, debug=0):
        """Add indirectly referenced macros and variables, adding them to the defines dictionary.

        Indirectly referenced macros and variables are macros and variables embedded in existing
        variables, source macros and include lists. For example, SIMHD is an indirect reference
        in "KA10D = ${SIMHD}/ka10" because KA10D might never have been expanded by 'extract()'.
        """

        def scan_var(varset, var):
            tmp = var
            return varset.union(set(SPM.extract_variables(tmp)))

        def replace_simhd(l, v):
            l.append(v.replace('SIMHD', 'CMAKE_SOURCE_DIR'))
            return l

        simvars = set()
        for v in self.vars.values():
            if isinstance(v, list):
                simvars = functools.reduce(scan_var, v, simvars)
            else:
                simvars = scan_var(simvars, v)

        for dir in self.dirs.keys():
            simvars = simvars.union(self.dirs[dir].get_simulator_vars(debug))

        if debug >= 2:
            print('Collected simvars:')
            pprint.pprint(simvars)

        for var in simvars:
            if var not in self.vars:
                if var in defs:
                    self.vars[var] = defs[var]
                else:
                    print('{0}: variable not defined.'.format(var))

        ## Replace SIMHD with CMAKE_SOURCE_DIR
        for k, v in self.vars.items():
            if isinstance(v, list):
                v = functools.reduce(replace_simhd, v, [])
            else:
                v = v.replace('SIMHD', 'CMAKE_SOURCE_DIR')
            self.vars[k] = v

    def collect_source_macros(self, comp, defs, varname, simcoll, sim, debug=0, depth=''):
        def inner_sources(srcmacro):
            for v in srcmacro:
                if not self.source_elems_to_ignore(v):
                    m = SPM._var_rx.match(v)
                    if m is not None:
                        vname = m.group(1)
                        vardef = defs.get(vname)
                        if vardef is not None:
                            ## Convert the macro variable into a list
                            vardef = [ SPM.normalize_variables(v) \
                                       for v in vardef.split() if not self.source_elems_to_ignore(v) ]
                            SU.emit_debug(debug, 3, '{0}source macro: {1} -> {2}'.format(depth, vname, vardef))
                            simcoll.add_source_macro(vname, vardef, sim)
                            ## Continue into the macro variable's definitions
                            inner_sources(vardef)

        vardef = defs.get(varname)
        if vardef is not None:
            vardef = [ SPM.normalize_variables(v) \
                       for v in vardef.split() if not self.source_elems_to_ignore(v) ]
            SU.emit_debug(debug, 3, '{0}source macro: {1} -> {2}'.format(depth, varname, vardef))
            simcoll.add_source_macro(varname, vardef, sim)
            inner_sources(vardef)
            SU.emit_debug(debug, 3, '{0}source added: {1}'.format(depth, comp))
            sim.add_source(comp)
        else:
            print('{0}undefined make macro: {1}'.format(depth, varname))

    def write_vars(self, stream):
        def collect_vars(varlist, var):
            varlist.extend(SPM.extract_variables(var))
            return varlist

        varnames = list(self.vars.keys())
        namewidth = max(map(lambda s: len(s), varnames))
        # vardeps maps the parent variable to its dependents, e.g.,
        # INTELSYSD -> [ISYS8010D, ...]
        vardeps = dict()
        # alldeps is the set of all parent and dependents, which will be
        # deleted from a copy of self.vars. The copy has variables that
        # don't depend on anything (except for CMAKE_SOURCE_DIR, but we
        # know that's defined by CMake.)
        alldeps = set()
        for var in varnames:
            if isinstance(self.vars[var], list):
                mvars = functools.reduce(collect_vars, self.vars[var], [])
            else:
                mvars = SPM.extract_variables(self.vars[var])
            mvars = [mvar for mvar in mvars if mvar != "CMAKE_SOURCE_DIR"]
            if mvars:
                alldeps.add(var)
                for mvar in mvars:
                    if mvar not in vardeps:
                        vardeps[mvar] = []
                    vardeps[mvar].append(var)
                    alldeps.add(mvar)

        nodeps = self.vars.copy()
        ## SIMHD will never be used.
        if 'SIMHD' in nodeps:
            del nodeps['SIMHD']
        for dep in alldeps:
            del nodeps[dep]

        varnames = list(nodeps.keys())
        varnames.sort()
        for var in varnames:
            self.emit_value(var, self.vars[var], stream, namewidth)

        ## Now to emit the dependencies
        depnames = list(vardeps.keys())
        depnames.sort()
        for dep in depnames:
            self.write_dep(dep, vardeps, alldeps, namewidth, stream)

        ## stream.write('\n## ' + '-' * 40 + '\n')

    def write_dep(self, dep, vardeps, alldeps, width, stream):
        # Not the most efficient, but it works
        alldeps.discard(dep)
        for parent in [ v for v in vardeps.keys() if dep in vardeps[v]]:
            if dep in parent.values():
                self.write_dep(parent, vardeps, alldeps, width, stream)

        stream.write('\n')
        self.emit_value(dep, self.vars[dep], stream, width)

        children = vardeps[dep]
        children.sort()
        for child in children:
            if child in alldeps:
                self.emit_value(child, self.vars[child], stream, width)
        alldeps -= set(children)

    def emit_value(self, var, value, stream, width=0):
        if isinstance(value, list):
            stream.write('set({:{width}} {})\n'.format(var, ' '.join(map(lambda s: '"' + s + '"', value)),
                                                       width=width))
        else:
            stream.write('set({:{width}} "{}")\n'.format(var, value, width=width))

    def write_simulators(self, toplevel_dir, debug=0):
        dirnames = list(self.dirs.keys())
        dirnames.sort()

        for subdir in dirnames:
            simcoll = self.dirs[subdir]
            test_label = subdir
            # Group tests under subdirectories together
            has_slash = test_label.find('/')
            if has_slash < 0:
                has_slash = test_label.find('\\')
            if has_slash >= 0:
                test_label = test_label[:has_slash]
            ## Write out individual CMakeLists.txt:
            subdir_cmake = os.path.join(toplevel_dir, subdir, 'CMakeLists.txt')
            print('==== writing to {0}'.format(subdir_cmake))
            with open(subdir_cmake, "w", newline='\r\n') as stream2:
                plural = '' if len(self.dirs[subdir]) == 1 else 's'
                stream2.write('## {} simulator{plural}\n'.format(subdir, plural=plural))
                stream2.write('\n'.join(_individual_header))
                stream2.write('\n')
                stream2.write('if (HAVE_UNITY_FRAMEWORK AND EXISTS "${CMAKE_CURRENT_SOURCE_DIR}/unit-tests/CMakeLists.txt")\n')
                stream2.write('  add_subdirectory(unit-tests)\n')
                stream2.write('endif ()')
                stream2.write('\n')
                simcoll.write_simulators(stream2, debug=debug, test_label=test_label)

            ## Write out unit tests, if the unit test subdirectory exists:
            subdir_units_dir = os.path.join(toplevel_dir, subdir, 'unit-tests')
            subdir_units = os.path.join(subdir_units_dir, 'CMakeLists.txt')
            if os.path.exists(subdir_units):
                with open(subdir_units, "w") as stream2:
                    plural = '' if len(self.dirs[subdir]) == 1 else 's'
                    stream2.write('## {} simulator{plural}\n'.format(subdir, plural=plural))
                    stream2.write('\n'.join(_unit_test_header))
                    stream2.write('\n')
                    simcoll.write_unit_tests(stream2, debug, subdir)

        simh_subdirs = os.path.join(toplevel_dir, 'cmake', 'simh-simulators.cmake')
        print("==== writing {0}".format(simh_subdirs))
        with open(simh_subdirs, "w") as stream2:
            stream2.write('\n'.join(_individual_header))
            self.write_vars(stream2)
            stream2.write('\n## ' + '-' * 40 + '\n')
            stream2.write(_unset_display_vars)
            stream2.write('\n## ' + '-' * 40 + '\n\n')
            stmts = [ 'add_subdirectory(' + dir + ')' for dir in dirnames ]
            stream2.write('\n'.join(stmts))
            stream2.write('\n')

    ## Representation when printed
    def __repr__(self):
        return '{0}({1}, {2})'.format(self.__class__.__name__, self.dirs.__repr__(), self.vars.__repr__())


if '_dispatch' in pprint.PrettyPrinter.__dict__:
    def cmake_pprinter(pprinter, cmake, stream, indent, allowance, context, level):
        cls = cmake.__class__
        stream.write(cls.__name__ + '(')
        indent += len(cls.__name__) + 1
        pprinter._format(cmake.dirs, stream, indent, allowance + 2, context, level)
        stream.write(',\n' + ' ' * indent)
        pprinter._format(cmake.vars, stream, indent, allowance + 2, context, level)
        stream.write(')')

    pprint.PrettyPrinter._dispatch[CMakeBuildSystem.__repr__] = cmake_pprinter



## generate.py
##
## Generate the simulator CMakeLists.txt from the top-level makefile.
##
## This is the top-level driver: process options, search for the
## makefile, parse the makefile and walk its dependencies, and,
## finally, output the CMakeLists.txt(s) and simh-simulators.cmake.
##
## Author: B. Scott Michel
## ("scooter me fecit")

import sys
import os.path
import argparse
import re

GEN_SCRIPT_DIR = os.path.dirname(__file__)
GEN_SCRIPT_NAME = os.path.basename(__file__)

import pprint

import simgen.cmake_container as SCC
import simgen.parse_makefile as SPM
import simgen.packaging as SPKG


def process_makefile(makefile_dir, debug=0):
    the_makefile = os.path.join(makefile_dir, "makefile")
    print('{0}: Processing {1}'.format(GEN_SCRIPT_NAME, the_makefile))

    (defs, rules, actions) = SPM.parse_makefile(the_makefile)
    if debug >= 4:
        pprint.pp(defs)

    all_rule = rules.get('all') or rules.get('ALL')
    if all_rule is None:
        print('{0}: "all" rule not found. Cannot proceed.'.format(GEN_SCRIPT_NAME))

    simulators = SCC.CMakeBuildSystem()
    for all_targ in SPM.shallow_expand_vars(all_rule, defs).split():
        print("{0}: all target {1}".format(GEN_SCRIPT_NAME, all_targ))
        walk_target_deps(all_targ, defs, rules, actions, simulators, debug=debug)

    experimental_rule = rules.get('experimental')
    if experimental_rule is not None:
        for experimental_targ in SPM.shallow_expand_vars(experimental_rule, defs).split():
            print("{0}: exp target {1}".format(GEN_SCRIPT_NAME, experimental_targ))
            walk_target_deps(experimental_targ, defs, rules, actions, simulators, debug=debug)

    simulators.collect_vars(defs, debug=debug)
    return simulators


## Makefile target dependencies to filter out.
_ignored_deps = [
    '${SIM}',
    '${BUILD_ROMS}'
]

## Simulator compile/link action pattern
_compile_act_rx = re.compile(r"\$[({]CC[)}]\s*(.*)")
_test_name_rx    = re.compile(r"\$@\s*\$\(call\s+find_test,\s*(.*),(.*)\)\s+\$")

def walk_target_deps(target, defs, rules, actions, simulators, depth='', debug=0):
    """ Recursively walk a target's dependencies, i.e., the right hand side of a make rule.
    Descend into each dependency to find something that looks like a simulator's
    source code list. Once source code list is found, extract simulator defines, includes,
    source files and set flags.
    """
    if debug >= 1:
        print('{0}-- target: {1}'.format(depth, target))

    target_deps = SPM.target_dep_list(target, rules, defs)
    if debug >= 2:
        print('{0}target deps: {1}'.format(depth, target_deps))

    has_buildrom = any(filter(lambda dep: dep == '${BUILD_ROMS}', target_deps))
    if debug >= 1:
        print('{0}   has_buildrom {1}', has_buildrom)

    deps = [dep for dep in target_deps if dep not in _ignored_deps]
    targ_actions = actions.get(target)
    if targ_actions:
        depth3 = depth + '   '
        if debug >= 2:
            print('{0}deps {1}'.format(depth3, deps))

        # Are the dependencies a source code list?
        expanded_deps = [l for slist in [ SPM.shallow_expand_vars(dep, defs).split() for dep in deps ] for l in slist]
        if debug >= 3:
            print('{0}expanded_deps {1}'.format(depth3, expanded_deps))

        if any(filter(lambda f: f.endswith('.c'), expanded_deps)):
            if debug >= 1:
                print('{0}sim sources {1}'.format(depth3, deps))
            if debug >= 2:
                print('{0}targ_actions {1}'.format(depth3, targ_actions))

            # The simulators' compile and test actions are very regular and easy to find:
            compile_act = None
            test_name   = None
            sim_dir     = None
            for act in targ_actions:
                m_cact = _compile_act_rx.match(act)
                m_test = _test_name_rx.match(act)
                if m_cact:
                    compile_act = m_cact.group(1)
                elif m_test:
                    (sim_dir, test_name) = m_test.group(1, 2)

            if debug >= 2:
                print('{0}sim_dir     {1}'.format(depth3, sim_dir))
                print('{0}compile_act {1}'.format(depth3, compile_act))
                print('{0}test_name   {1}'.format(depth3, test_name))

            if compile_act and test_name and sim_dir:
                sim_name = target.replace("${BIN}", "").replace("${EXE}", "")
                # Just in case there are vestiges of old-style make variables
                sim_name = sim_name.replace("$(BIN)", "").replace("$(EXE)", "")
                if debug >= 2:
                    print('{0}sim_name    {1}'.format(depth3, sim_name))

                simulators.extract(compile_act, test_name, sim_dir, sim_name, defs, has_buildrom, debug, depth+'  ')
    else:
        # No actions associated with the dependency(ies), which means that the dependency(ies)
        # are meta-targets. Continue to walk.
        for dep in deps:
            walk_target_deps(dep, defs, rules, actions, simulators, depth=depth+'  ', debug=debug)


if __name__ == '__main__':
    args = argparse.ArgumentParser(description="SIMH simulator CMakeLists.txt generator.")
    args.add_argument('--debug', nargs='?', const=1, default=0, type=int,
                      help='Debug level (0-3, 0 == off)')
    args.add_argument('--srcdir', default=None,
                      help='makefile source directory.')
    args.add_argument('--skip-orphans', action='store_true',
                      help='Skip the check for packaging orphans')

    flags = vars(args.parse_args())

    debug_level = flags.get('debug')
    makefile_dir = flags.get('srcdir')

    found_makefile = True
    if makefile_dir is None:
        ## Find the makefile, which should be one directory up from this Python
        ## module
        makefile_dir = GEN_SCRIPT_DIR
        print('{0}: Looking for makefile, starting in {1}'.format(GEN_SCRIPT_NAME, makefile_dir))
        the_makefile = ''
        while makefile_dir:
            the_makefile = os.path.join(makefile_dir, "makefile")
            if os.path.exists(the_makefile):
                break
            else:
                makefile_dir = os.path.dirname(makefile_dir)
                print('{0}: Looking for makefile, trying {1}'.format(GEN_SCRIPT_NAME, makefile_dir))

        if not the_makefile:
            found_makefile = False
    else:
        the_makefile = os.path.join(makefile_dir, "makefile")
        if not os.path.exists(the_makefile):
            found_makefile = False

    if not found_makefile:
        print('{0}: SIMH top-level makefile not found, relative to {1}'.format(GEN_SCRIPT_NAME, GEN_SCRIPT_DIR))
        sys.exit(1)

    sims = process_makefile(makefile_dir, debug=debug_level)

    ## Sanity check: Make sure that all of the simulators in SPKG.package_info have
    ## been encountered
    for simdir in sims.dirs.keys():
        if debug_level >= 2:
            print('{}: simdir {}'.format(GEN_SCRIPT_NAME, simdir))
        for sim in sims.dirs[simdir].simulators.keys():
            SPKG.package_info[sim].encountered()

    if not flags.get('skip_orphans'):
        print('{0}: Expecting to emit {1} simulators.'.format(GEN_SCRIPT_NAME, len(SPKG.package_info.keys())))

        orphans = [ sim for sim, pkg_info in SPKG.package_info.items() if not pkg_info.was_processed() ]
        if len(orphans) > 0:
            print('{0}: Simulators not extracted from makefile:'.format(GEN_SCRIPT_NAME))
            for orphan in orphans:
                print('{0}{1}'.format(' ' * 4, orphan))
            sys.exit(1)
        else:
            print('{0}: All simulators present and accounted for!'.format(GEN_SCRIPT_NAME))

    if debug_level >= 1:
        pp = pprint.PrettyPrinter()
        pp.pprint(sims)

    ## Emit all of the individual CMakeLists.txt
    sims.write_simulators(makefile_dir, debug=debug_level)
    ## Emit the packaging data
    SPKG.write_packaging(makefile_dir)



"""Makefile parsing and variable expansion.

Read and collect variable, rule and action information from a [Mm]akefile.
This isn't a precise collection; for example, it does not respect GNU Makefile
directives such as 'ifeq' and 'ifneq'.
"""

import re

# Regexes needed for parsing Makefile (and similar syntaxes,
# like old-style Setup files).
_variable_rx = re.compile(r"\s*([A-Za-z][\w_-]+)\s*=\s*(.*)")
_rule_rx     = re.compile(r"(((\$[({])*\w[\w_-]+[)}]*)+)\s*:\s*(.*)")

# Regex that recognizes variables. Group 1 is the variable's name.
_var_pattern = r"[A-Za-z][\w_-]*"
_var_rx      = re.compile(r"^\$[{(](" + _var_pattern + r")[)}]$")
_var_rx2     = re.compile(r"\$[{(]("  + _var_pattern + r")[)}]")
_norm_var_rx = re.compile(r"\$[(]("   + _var_pattern + r")[)]")

def parse_makefile(fn, g_vars=None, g_rules=None, g_actions=None):
    """Parse a Makefile-style file.

    Collects all of the variable definitions, rules and actions associated with rules.

    """
    ## Python 3.11 and onward dropped distuitls, so import our local copy.
    from simgen.text_file import TextFile

    fp = TextFile(fn, strip_comments=1, skip_blanks=1, join_lines=1, errors="surrogateescape")

    if g_vars is None:
        g_vars = {}
    if g_rules is None:
        g_rules = {}
    if g_actions is None:
        g_actions = {}
    done = {}
    rules = {}
    actions = {}

    line = fp.readline()
    while line is not None:
        vmatch = _variable_rx.match(line)
        rmatch = _rule_rx.match(line)
        if vmatch:
            n, v = vmatch.group(1, 2)
            v = v.strip()

            try:
                v = int(v)
            except ValueError:
                # insert literal `$'
                done[n] = v.replace('$$', '$')
            else:
                done[n] = v

            line = fp.readline()
        elif rmatch:
            ## make allows "$(VAR)" and "${VAR}" to be used interchangably, so there's a
            ## possibility that the sim developer used both forms in the target, e.g.:
            ##
            ## foosim: $(BIN)foosim$(EXE)
            ##
            ## ${BIN}foosim${EXE}: stuff that foosim depends on...

            n, v = rmatch.group(1, 4)
            rules[normalize_variables(n)] = normalize_variables(v)

            ## Collect the actions:
            collected = []
            line = fp.readline()
            while line is not None:
                m = _variable_rx.match(line) or _rule_rx.match(line)
                if m is None:
                    collected.append(line.lstrip())
                    line = fp.readline()
                else:
                    break
            actions[n] = collected
        else:
            line = fp.readline()

    fp.close()

    # strip spurious spaces
    for k, v in done.items():
        if isinstance(v, str):
            done[k] = v.strip().replace('\t', ' ')

    # save the results in the global dictionary
    g_vars.update(done)
    g_rules.update(rules)
    g_actions.update(actions)
    return (g_vars, g_rules, g_actions)


def target_dep_list(target, rules, defs):
    return (rules.get(target) or '').split()

def expand_vars(s, defs):
    """Expand Makefile-style variables -- "${foo}" or "$(foo)" -- in
    'string' according to 'defs' (a dictionary mapping variable names to
    values).  Variables not present in 'defs' are silently expanded to the
    empty string.

    Returns a variable-expanded version of 's'.
    """

    # This algorithm does multiple expansion, so if defs['foo'] contains
    # "${bar}", it will expand ${foo} to ${bar}, and then expand
    # ${bar}... and so forth.  This is fine as long as 'defs' comes from
    # 'parse_makefile()', which takes care of such expansions eagerly,
    # according to make's variable expansion semantics.

    while True:
        m = _var_rx2.search(s)
        if m:
            (beg, end) = m.span()
            s = s[0:beg] + (defs.get(m.group(1)) or '') + s[end:]
        else:
            break
    return s


def shallow_expand_vars(s, defs):
    """Expand Makefile-style variables -- "${foo}" or "$(foo)" -- in
    'string' according to 'defs' (a dictionary mapping variable names to
    values).  Variables not present in 'defs' are silently expanded to the
    empty string.

    Returns a variable-expanded version of 's'.
    """

    # This algorithm does multiple expansion, so if defs['foo'] contains
    # "${bar}", it will expand ${foo} to ${bar}, and then expand
    # ${bar}... and so forth.  This is fine as long as 'defs' comes from
    # 'parse_makefile()', which takes care of such expansions eagerly,
    # according to make's variable expansion semantics.

    m = _var_rx2.search(s)
    if m:
        (beg, end) = m.span()
        return s[0:beg] + (defs.get(m.group(1)) or '') + shallow_expand_vars(s[end:], defs)

    return s


def extract_variables(varstr):
    """Extracct all variable references, e.g., "${foo}" or "$(foo)"
    from a string.
    """
    retval = []
    tmp = varstr
    while True:
        m = _var_rx2.search(tmp)
        if m:
            retval.append(m[1])
            tmp = tmp[m.end():]
        else:
            break
    return retval


def normalize_variables(varstr):
    """Convert '$(var)' to '${var}' -- normalizes all variables to a consistent
    form.
    """
    retval = ""
    tmp = varstr
    while tmp:
        m = _norm_var_rx.search(tmp)
        if m:
            retval += tmp[:m.start()] + "${" + m[1] + "}"
            tmp = tmp[m.end():]
        else:
            retval += tmp
            tmp = ""
    return retval


def test_rule_rx():
    result = _rule_rx.match('${BIN}frontpaneltest${EXE} : frontpanel/FrontPanelTest.c sim_sock.c sim_frontpanel.c')
    print('{0}: {1}'.format('${BIN}frontpaneltest${EXE}...', result))
    print(result.groups())


def test_normalize_variables():
    result = normalize_variables('foo: bar baz')
    print('{0}: {1}'.format('foo:...', result))
    result = normalize_variables('$(var): dep1 dep2')
    print('{0}: {1}'.format('$(var)...', result))
    result = normalize_variables('$(var): dep1 ${var2} dep2 $(var3)')
    print('{0}: {1}'.format('$(var)...', result))



## KA10 simulator: Add the PANDA_LIGHTS and the PIDP10 options for frontpanel
## code.

import simgen.basic_simulator as SBS

class KA10Simulator(SBS.SIMHBasicSimulator):
    def __init__(self, sim_name, dir_macro, test_name, buildrom):
        super().__init__(sim_name, dir_macro, test_name, buildrom)

    def write_simulator(self, stream, indent, test_label='ka10'):
        ## Keep the CMake options separate, just in case they are needed for more
        ## than just the KA10 (and add a class variable so they are only written
        ## once.)
        stream.write('''
option(PANDA_LIGHTS
       "Enable (=1)/disable (=0) KA-10/KI-11 simulator\'s Panda display. (def: disabled)"
       FALSE)
option(PIDP10
       "Enable (=1)/disable (=0) PIDP10 display options (def: disabled)"
       FALSE)

### Ensure that the options are mutually exclusive:
if (PANDA_LIGHTS AND PIDP10)
  message(FATAL_ERROR "PANDA_LIGHTS and PIDP10 options are mutually exclusive. Choose one.")
endif ()


''')
        ## Emit the simulator:
        super().write_simulator(stream, indent, test_label)
        ## Update the display sources for PANDA_LIGHTS or PIDP10:
        stream.write('\n'.join([
            '',
            'if (PANDA_LIGHTS)',
            '  target_sources({0} PUBLIC {1}/kx10_lights.c)'.format(self.sim_name, self.dir_macro),
            '  target_compile_definitions({0} PUBLIC PANDA_LIGHTS)'.format(self.sim_name),
            '  target_link_libraries({0} PUBLIC usb-1.0)'.format(self.sim_name),
            'endif ()',
            'if (PIDP10)',
            '  target_sources({0} PUBLIC {1}/ka10_pipanel.c)'.format(self.sim_name, self.dir_macro),
            '  target_compile_definitions({0} PUBLIC PIDP10=1)'.format(self.sim_name),
            'endif ()'
        ]))
        stream.write('\n')




## IBM 1130 simulator customizations:
##
## - Add the Win32 resource file for Windows builds
## - Add the "-g" test flag to bypass/disable the simulator GUI when
##   running RegisterSanityCheck test.
import simgen.basic_simulator as SBS

class IBM1130Simulator(SBS.SIMHBasicSimulator):
    '''The IBM650 simulator creates relatively deep stacks, which will fail on Windows.
    Adjust target simulator link flags to provide a 8M stack, similar to Linux.
    '''
    def __init__(self, sim_name, dir_macro, test_name, buildrom):
        super().__init__(sim_name, dir_macro, test_name, buildrom, test_args="-g")

    def write_simulator(self, stream, indent, test_label='ibm650'):
        super().write_simulator(stream, indent, test_label)
        stream.write('\n'.join([
            '',
            'if (WIN32)',
            '    ## Add GUI support, compile in resources:',
            '    target_compile_definitions(ibm1130 PRIVATE GUI_SUPPORT)',
            '    target_sources(ibm1130 PRIVATE ibm1130.rc)',
            'endif()',
            '',
            '# IBM 1130 utilities:',
            '# add_subdirectory(utils)',
            ''
        ]))





## VAX simulators require extra magic -- notably, 'microvax3900${EXE}' needs
## to be symlinked, hardlinked or copied (in that order) to 'vax${EXE}'.

import simgen.basic_simulator as SBS

class BasicVAXSimulator(SBS.SIMHBasicSimulator):
    """
    """
    def __init__(self, sim_name, dir_macro, test_name, buildrom):
        super().__init__(sim_name, dir_macro, test_name, buildrom)

    def write_unit_test(self, stream, indent, individual=False, test_label='default'):
        stream.write('\n')
        self.write_section(stream, 'add_unit_test', indent, individual=False, test_label=test_label,
                           section_name='vax_cc_{}'.format(self.sim_name),
                           section_srcs=['vax_cc.c'],
                           section_incs=self.includes)

class VAXSimulator(BasicVAXSimulator):
    """
    """
    def __init__(self, sim_name, dir_macro, test_name, buildrom):
        super().__init__(sim_name, dir_macro, test_name, buildrom)

    def write_simulator(self, stream, indent, test_label='VAX'):
        super().write_simulator(stream, indent, test_label)
        stream.write('''
set(vax_binary_dir ${CMAKE_RUNTIME_OUTPUT_DIRECTORY})
if (CMAKE_CONFIGURATION_TYPES)
    string(APPEND vax_binary_dir "/$<CONFIG>")
endif (CMAKE_CONFIGURATION_TYPES)

add_custom_command(TARGET vax POST_BUILD
    COMMAND "${CMAKE_COMMAND}"
        -DSRCFILE=vax${CMAKE_EXECUTABLE_SUFFIX}
        -DDSTFILE=microvax3900${CMAKE_EXECUTABLE_SUFFIX}
        -DWORKING_DIR=${vax_binary_dir}
        -P ${CMAKE_SOURCE_DIR}/cmake/file-link-copy.cmake
    COMMENT "Symlink vax${CMAKE_EXECUTABLE_SUFFIX} to microvax3900${CMAKE_EXECUTABLE_SUFFIX}"
    WORKING_DIRECTORY ${vax_binary_dir})

install(
    CODE "
        execute_process(
            COMMAND ${CMAKE_COMMAND}
                -DSRCFILE=vax${CMAKE_EXECUTABLE_SUFFIX}
                -DDSTFILE=microvax3900${CMAKE_EXECUTABLE_SUFFIX}
                -DWORKING_DIR=\$ENV{DESTDIR}\${CMAKE_INSTALL_PREFIX}/bin
                -P ${CMAKE_SOURCE_DIR}/cmake/file-link-copy.cmake)"
    COMPONENT vax_family)
''')
        stream.write('\n')




import pprint

import simgen.parse_makefile as SPM
import simgen.utils as SU
import simgen.packaging as SPKG

class SIMHBasicSimulator:
    """
    """
    def __init__(self, sim_name, dir_macro, test_name, buildrom, test_args=None):
        self.sim_name = sim_name
        ## self.dir_macro     -> Directory macro (e.g., "${PDP11D}" for source
        self.dir_macro = dir_macro
        self.test_name = test_name
        self.int64 = False
        self.full64 = False
        self.buildrom = buildrom
        ## self.has_display    -> True if there is a specific display used by the simulator.
        self.has_display = False
        ## self.uses_video     -> True if USE_SIM_VIDEO appears in the simulator's preprocessor defn's.
        self.uses_video = False
        ## self.besm6_sdl_hack -> Only set/used by the BESM6 simulator.
        self.besm6_sdl_hack = False
        ## self.uses_aio       -> True if the simulator uses AIO
        self.uses_aio = False
        ## self.test_args      -> Simulator flags to pass to the test phase. Used by ibm1130 to
        ##                        pass "-g" to disable to the GUI. This argument can be a single
        ##                        string or a list.
        self.test_args = test_args

        self.sources = []
        self.defines = []
        self.includes = []

    def add_source(self, src):
        if src not in self.sources:
            self.sources.append(src)

    def add_include(self, incl):
        if incl not in self.includes:
            self.includes.append(incl)

    def add_define(self, define):
        if define not in self.defines:
            self.defines.append(define)

    def scan_for_flags(self, defs):
        """Scan for USE_INT64/USE_ADDR64 in the simulator's defines and set the
        'int64' and 'full64' instance variables. If found, these defines are
        removed. Also look for any of the "DISPLAY" make macros, and, if found,
        set the 'video' instance variable.
        """
        use_int64 = 'USE_INT64' in self.defines
        use_addr64 = 'USE_ADDR64' in self.defines
        if use_int64 or use_addr64:
            self.int64  = use_int64 and not use_addr64
            self.full64 = use_int64 and use_addr64
            for defn in ['USE_INT64', 'USE_ADDR64']:
                try:
                    self.defines.remove(defn)
                except:
                    pass

        ## Video support:

        self.has_display = any(map(lambda s: 'DISPLAY' in SPM.shallow_expand_vars(s, defs), self.sources))
        if self.has_display:
            for src in ['${DISPLAYL}', '$(DISPLAYL)']:
                try:
                    self.sources.remove(src)
                except:
                    pass

        self.uses_video = 'USE_SIM_VIDEO' in self.defines or self.has_display

        ## AIO support:
        self.uses_aio   = 'SIM_ASYNCH_IO' in self.defines
        if self.uses_aio:
            for defn in ['SIM_ASYNCH_IO', 'USE_READER_THREAD']:
                try:
                    self.defines.remove(defn)
                except:
                    pass

    def cleanup_defines(self):
        """Remove command line defines that aren't needed (because the CMake interface libraries
        already define them.)
        """
        for define in ['USE_SIM_CARD', 'USE_SIM_VIDEO', 'USE_NETWORK', 'USE_SHARED']:
            try:
                self.defines.remove(define)
            except:
                pass

    def get_source_vars(self):
        srcvars = set()
        for src in self.sources:
            srcvars = srcvars.union(set(SPM.extract_variables(src)))
        return srcvars

    def get_include_vars(self):
        incvars = set()
        for inc in self.includes:
            incvars = incvars.union(set(SPM.extract_variables(inc)))
        return incvars

    def write_section(self, stream, section, indent, test_label='default', additional_text=[],
                      section_name=None, section_srcs=None, section_incs=None):
        indent4 = ' ' * (indent + 4)
        indent8 = ' ' * (indent + 8)

        pkg_info = SPKG.package_info.get(section_name)
        install_flag = pkg_info.install_flag if pkg_info is not None else None
        pkg_family = pkg_info.family.component_name if pkg_info is not None else None

        stream.write(' ' * indent + '{}({}\n'.format(section, section_name))
        stream.write(' ' * (indent + 4) + 'SOURCES\n')
        stream.write('\n'.join(map(lambda src: indent8 + src, section_srcs)))
        if len(self.includes) > 0:
            stream.write('\n' + indent4 + 'INCLUDES\n')
            stream.write('\n'.join([ indent8 + inc for inc in section_incs]))
        if len(self.defines) > 0:
            stream.write('\n' + indent4 + 'DEFINES\n')
            stream.write('\n'.join(map(lambda dfn: indent8 + dfn, self.defines)))
        if self.int64:
            stream.write('\n' + indent4 + 'FEATURE_INT64')
        if self.full64:
            stream.write('\n' + indent4 + 'FEATURE_FULL64')
        if self.uses_video:
            stream.write('\n' + indent4 + "FEATURE_VIDEO")
        if self.has_display:
            stream.write('\n' + indent4 + "FEATURE_DISPLAY")
        if self.besm6_sdl_hack:
            stream.write('\n' + indent4 + "BESM6_SDL_HACK")
        if self.uses_aio:
            stream.write('\n' + indent4 + "USES_AIO")
        if self.test_args:
            out_args = self.test_args
            if isinstance(self.test_args, str):
                out_args = self.test_args.split()

            out_args = ' '.join('"{0}"'.format(w) for w in out_args)
            stream.write('\n' + indent4 + 'TEST_ARGS {}'.format(out_args))
        if self.buildrom:
            stream.write('\n' + indent4 + "BUILDROMS")
        stream.write('\n' + indent4 + "LABEL " + test_label)
        if install_flag:
            if pkg_family:
                stream.write('\n' + indent4 + "PKG_FAMILY " + pkg_family)
        else:
            stream.write('\n' + indent4 + "NO_INSTALL")
        stream.write('\n' + '\n'.join(additional_text))
        stream.write(')\n')

    def write_simulator(self, stream, indent, test_label='default'):
        ## When writing an individual CMakeList.txt, we can take advantage of the CMAKE_CURRENT_SOURCE_DIR
        ## as a replacement for the SIMH directory macro.
        srcs = [ src.replace(self.dir_macro + '/', '') for src in self.sources]
        incs = [ inc if inc != self.dir_macro else '${CMAKE_CURRENT_SOURCE_DIR}' for inc in self.includes]

        indent4 = ' ' * (indent + 4)

        addl_text = [
            indent4 + "TEST " + self.test_name,
        ]

        self.write_section(stream, 'add_simulator', indent, test_label, additional_text=addl_text,
                           section_name=self.sim_name, section_srcs=srcs, section_incs=incs)

    # Default: Don't generate a unit test CMakeFiles.txt. Yet.
    def write_unit_test(self, stream, indent, test_label='default'):
        pass

    def __repr__(self):
        return '{0}({1},{2},{3},{4})'.format(self.__class__.__name__, self.sim_name.__repr__(),
            self.sources.__repr__(), self.includes.__repr__(), self.defines.__repr__())


class BESM6Simulator(SIMHBasicSimulator):
    """The (fine Communist) BESM6 simulator needs some extra code
    in the CMakeLists.txt to detect a suitable font that supports
    Cyrillic.
    """
    def __init__(self, sim_name, dir_macro, test_name, buildrom):
        super().__init__(sim_name, dir_macro, test_name, buildrom)

    def scan_for_flags(self, defs):
        super().scan_for_flags(defs)

    def write_simulator(self, stream, indent, test_label='besm6'):
        ## Fixups... :-)
        for macro in ['FONTFILE=$(FONTFILE)', 'FONTFILE=${FONTFILE}']:
            try:
                self.defines.remove(macro)
            except:
                pass

        ## Add the search for a font file.
        stream.write('\n'.join([
          'set(besm6_font)',
          'set(cand_fonts',
          '      "DejaVuSans.ttf"',
          '      "LucidaSansRegular.ttf"',
          '      "FreeSans.ttf"',
          '      "AppleGothic.ttf"',
          '      "tahoma.ttf")',
          'set(cand_fontdirs',
          '      "/usr/share/fonts"',
          '      "/usr/lib/jvm"',
          '      "/Library/Fonts"',
          '      "/System/Library/Fonts"',
          '      "/System/Library/Frameworks/JavaVM.framework/Versions"',
          '      "$ENV{WINDIR}/Fonts")',
          '',
          'foreach (fdir ${cand_fontdirs})',
          '    foreach (font ${cand_fonts})',
          '        if (EXISTS ${fdir}/${font})',
          '            get_filename_component(fontfile ${fdir}/${font} ABSOLUTE)',
          '            list(APPEND besm6_font ${fontfile})',
          '        endif ()',
          '',
          '        file(GLOB besm6_font_cand_1 LIST_DIRECTORIES FALSE "${fdir}/*/${font}")',
          '        file(GLOB besm6_font_cand_2 LIST_DIRECTORIES FALSE "${fdir}/*/*/${font}")',
          '        file(GLOB besm6_font_cand_3 LIST_DIRECTORIES FALSE "${fdir}/*/*/*/${font}")',
          '        list(APPEND besm6_font ${besm6_font_cand_1} ${besm6_font_cand_2} ${besm6_font_cand_3})',
          '    endforeach()',
          'endforeach()',
          '',
          'if (besm6_font)',
          '    set(besm6_found_fonts "BESM6: Fonts found")',
          '    foreach(bfont ${besm6_font})',
          '        string(APPEND besm6_found_fonts "\n      .. ${bfont}")',
          '    endforeach ()',
          '    message(STATUS ${besm6_found_fonts})',
          '    unset(besm6_found_fonts)',
          '    list(GET besm6_font 0 besm6_font)',
          '    message(STATUS "BESM6: Using ${besm6_font}")',
          'else ()',
          '    set(besm6_no_fonts "BESM6: No applicable Cyrillic fonts found.")',
          '    string(APPEND besm6_no_fonts "\n    Font names tried:")',
          '    foreach (font ${cand_fonts})',
          '        string(APPEND besm6_no_fonts "\n      ..  ${font}")',
          '    endforeach ()',
          '    string(APPEND besm6_no_fonts "\n\n    Looked in:")',
          '    foreach (fdir ${cand_fontdirs})',
          '        string(APPEND besm6_no_fonts "\n      ..  ${fdir}")',
          '    endforeach()',
          '    string(APPEND besm6_no_fonts "\n\nBESM6: Not building with panel display.")',
          '    message(STATUS ${besm6_no_fonts})',
          '    unset(besm6_no_fonts)',
          'endif ()',
          '',
          'if (NOT (besm6_font AND WITH_VIDEO))\n']))
        super().write_simulator(stream, indent + 4, test_label)
        stream.write('else ()\n')
        self.defines.append("FONTFILE=${besm6_font}")
        self.has_display = True
        self.uses_video = True
        self.besm6_sdl_hack = True
        super().write_simulator(stream, indent + 4, test_label)
        stream.write('\n'.join([
            'endif()',
            'unset(cand_fonts)',
            'unset(cand_fontdirs)\n']))

class IBM650Simulator(SIMHBasicSimulator):
    '''The IBM650 simulator creates relatively deep stacks, which will fail on Windows.
    Adjust target simulator link flags to provide a 8M stack, similar to Linux.
    '''
    def __init__(self, sim_name, dir_macro, test_name, buildrom):
        super().__init__(sim_name, dir_macro, test_name, buildrom)
        self.stack_size = 8 * 1024 * 1024

    def write_simulator(self, stream, indent, test_label='ibm650'):
        super().write_simulator(stream, indent, test_label)
        stream.write('\n')
        ## Link i650 with a 8M stack on windows
        stream.write('\n'.join([
            'if (WIN32)',
            '    if (MSVC)',
            '        set(I650_STACK_FLAG "/STACK:{0}")'.format(self.stack_size),
            '    else ()',
            '        set(I650_STACK_FLAG "-Wl,--stack,{0}")'.format(self.stack_size),
            '    endif ()',
            '    if (CMAKE_VERSION VERSION_GREATER_EQUAL "3.13")',
            '        target_link_options({0} PUBLIC "${{I650_STACK_FLAG}}")'.format(self.sim_name),
            '    else ()',
            '        set_property(TARGET {0} LINK_FLAGS " ${{I650_STACK_FLAG}}")'.format(self.sim_name),
            '    endif ()',
            'endif()'
        ]))

if '_dispatch' in pprint.PrettyPrinter.__dict__:
    def sim_pprinter(pprinter, sim, stream, indent, allowance, context, level):
        cls = sim.__class__
        stream.write(cls.__name__ + '(')
        indent += len(cls.__name__) + 1
        pprinter._format(sim.sim_name, stream, indent, allowance + 2, context, level)
        stream.write(',')
        pprinter._format(sim.dir_macro, stream, indent, allowance + 2, context, level)
        stream.write(',')
        pprinter._format(sim.int64, stream, indent, allowance + 2, context, level)
        stream.write(',')
        pprinter._format(sim.full64, stream, indent, allowance + 2, context, level)
        stream.write(',')
        pprinter._format(sim.has_display, stream, indent, allowance + 2, context, level)
        stream.write(',')
        pprinter._format(sim.sources, stream, indent, allowance + 2, context, level)
        stream.write(',\n' + ' ' * indent)
        pprinter._format(sim.defines, stream, indent, allowance + 2, context, level)
        stream.write(',\n' + ' ' * indent)
        pprinter._format(sim.includes, stream, indent, allowance + 2, context, level)
        stream.write(')')

    pprint.PrettyPrinter._dispatch[SIMHBasicSimulator.__repr__] = sim_pprinter
